import socket
import threading
import traceback

ENCODING = "utf-8"

HOST = "127.0.0.1"
PORT = 8881

server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
server_socket.bind((HOST, PORT))

server_socket.listen()


def get_header(headers: str, header: str) -> str:
    """
    Returns the content of a given header.
    """
    lines = headers.split("\n")
    for line in lines:
        if (line.lower().startswith(header.lower())):
            return line.split(': ', maxsplit=1)[-1].replace("\r", "")

    return ""


def modify_response(headers: bytes, content: bytes) -> bytes:
    """
    Modifies a response from a server.
    """
    content_type = get_header(headers.decode(), "Content-Type").split(';')[0]

    if content_type == "text/plain" or content_type == "text/html":
        result = b""

        for line in content.split(b'\n'):
            if b'<img' in line:
                result += line + b'\n'
            else:
                line = line.replace(b"Smiley", b"Trolly")
                line = line.replace(b"Stockholm", bytes("Linköping", ENCODING))
                result += line + b'\n'

        old_length = get_header(headers.decode(), "Content-Length").encode()
        new_length = bytes(len(result))
        headers = headers.replace(b"Content-Length: " + old_length, b"Content-Length: " + new_length)

        return headers + result

    return headers + content


def modify_request(request: str) -> str:
    """
    Modifies a request from the client.
    """
    to_replace = "http://zebroid.ida.liu.se/fakenews/smiley.jpg"
    new_img = "http://zebroid.ida.liu.se/fakenews/trolly.jpg"

    a = request.replace(to_replace, new_img)
    return a


def make_request(request: str) -> bytes:
    """
    Modifies and sends a request, and returns
    the modified response.
    """
    host_name = get_header(request, "Host")
    host = socket.gethostbyname(host_name)

    # Connect to server
    request_socket = socket.socket()
    request_socket.connect((host, 80))
    request_socket.settimeout(2)

    # Send request to server
    request_socket.send(modify_request(request).encode())

    # Get http headers
    headers, content = request_socket.recv(4096).split(b'\r\n\r\n', maxsplit=1)
    headers = headers + b'\r\n\r\n' 

    # If there is no content, return headers
    if not content:
        return headers

    content_length = int(get_header(headers.decode(ENCODING, "ignore"), "Content-Length"))

    # Get http content
    received = len(content)
    fragments = [content]
    while True: 
        try:
            chunk = request_socket.recv(4096)
            fragments.append(chunk)
            received += len(chunk)
            if not chunk or received >= content_length:
                break
        except: # If timed out
            break

    content = b''.join(fragments)

    return modify_response(headers, content)


def socket_handler(socket: socket):
    """
    Handles client sockets.
    """
    try:
        request = socket.recv(4096).decode(ENCODING, "ignore")
        print("Request from client: ", request)

        if is_valid_request(request):
            socket.send(make_request(request))

        socket.close()

    except:
        print("Socket error")
        traceback.print_exc()
        socket.close()


def is_valid_request(request: str) -> bool:
    """
    Returns true if a client request is valid.
    """
    return len(request) > 0


while True:
    try:
        # Wait for connection
        client_socket, client_addr = server_socket.accept()
        print("connected")

        thread = threading.Thread(target=socket_handler, args=[client_socket]) 
        thread.start()

        print(f"[ACTIVE CONNECTIONS] {threading.activeCount() - 1}")

    except:
        print("Threading error")
        traceback.print_exc()
        server_socket.close()
        client_socket.close()
        exit()
